from typing import List, Union
import typing
import aiohttp
import enum
from . import embed
from .user import User
from .guild import Member, Role
from .channel import ChannelMention
from pycordia import models

# class TextChannel:
#     pass

# class GuildMember:
#     pass


class MessageActivity:
    """A message activity object"""

    class ActivityTypes(enum.Enum):
        join = 1
        spectate = 2
        listen = 3
        join_request = 5

    def __init__(self, data: dict):
        Activity = MessageActivity.ActivityTypes
        self.activity_type = (
            None if (data.get("type") is None) else Activity(data.get("type"))
        )

        self.party_id: Union[str, None] = data.get("party_id")

    def __repr__(self):
        return f"<pycordia.models.MessageActivity - id={self.party_id} activity={self.activity_type}>"


class Application:
    def __init__(self, data: dict):
        self.app_id: Union[str, None] = data.get("id")
        self.name: Union[str, None] = data.get("name")
        self.icon: Union[str, None] = data.get("icon")
        self.description: Union[str, None] = data.get("description")

        self.rpc_origins: list = data.get("rpc_origins", [])
        self.bot_public: Union[bool, None] = data.get("bot_public")
        self.bot_require_code_grant: Union[bool, None] = data.get("bot_require")

        self.terms_of_service_url: Union[str, None] = data.get("terms_of_service_url")
        self.privacy_policy_url: Union[str, None] = data.get("privacy_policy_url")
        self.owner: User = User(data.get("owner", {}))

        self.cover_image_hash: Union[str, None] = data.get("cover_image")
        self.flags: Union[int, None] = data.get("flags")

    def __repr__(self):
        return f"<pycordia.models.Application - id={self.app_id} name={self.name}>"


class RoleTags:
    def __init__(self, data: dict):
        self.bot_id = data.get("bot_id")
        self.integration_id = data.get("integration_id")
        self.premium_subscriber = data.get("premium_subscriber")

    def __repr__(self):
        return f"<pycordia.models.RoleTags - id={self.bot_id}>"


class Role:
    def __init__(self, data: dict):
        self.role_id: Union[str, None] = data.get("id")
        self.role_name: Union[str, None] = data.get("name")

        self.color: Union[int, None] = data.get("color")
        self.colour = self.color

        self.hoist: Union[bool, None] = data.get("hoist")
        self.position: Union[int, None] = data.get("position")
        self.permissions: Union[str, None] = data.get("permissions")
        self.managed: Union[bool, None] = data.get("managed")
        self.mentionable: Union[bool, None] = data.get("mentionable")

        self.tags: Union[RoleTags, None] = data.get("tags")

    def __repr__(self):
        return f"<pycordia.models.Role - id={self.role_id} name={self.role_name}>"


class Reaction:
    def __init__(self, data: dict):
        self.count: Union[int, None] = data.get("count")
        self.was_me: Union[bool, None] = data.get("me")
        self.emoji: Emoji = Emoji(data.get("emoji", {}))

    def __repr__(self):
        return (
            f"<pycordia.models.Reaction - emoji={self.emoji.name} count={self.count}>"
        )


class Emoji:
    def __init__(self, data: dict):
        self.emoji_id = data.get("id")
        self.name = data.get("name")
        self.roles = [Role(role) for role in data.get("roles", [])]
        self.user = User(data.get("user", {}))
        self.requires_colons = data.get("require_colons")
        self.managed = data.get("managed")
        self.animated = data.get("animated")
        self.available = data.get("available")

    def __repr__(self):
        return f"<pycordia.models.Emoji - id={self.emoji_id} name={self.name}>"


class StickerItem:
    def __init__(self, data: dict):
        self.sticker_id = data.get("id")
        self.name = data.get("name")
        self.format_type = data.get("type")

    def __repr__(self):
        return f"<pycordia.models.StickerItem - id={self.sticker_id} name={self.name}>"


class Attachment:
    def __init__(self, data: dict):
        """An attachment for a message

        Arguments:
            attachment_id (int): The ID of the attachment
            filename (str): The attachment's filename
            content_type (str): The content type
            size (int):
            url (str):
            proxy_url (str):
            height (int):
            width (int):
        """
        self.attachment_id: Union[str, None] = data.get("id")
        self.filename: Union[str, None] = data.get("filename")
        self.content_type: Union[str, None] = data.get("content_type")
        self.size: Union[int, None] = data.get("size")
        self.url: Union[str, None] = data.get("url")
        self.proxy_url: Union[str, None] = data.get("proxy_url")
        self.height: Union[int, None] = data.get("height")
        self.width: Union[int, None] = data.get("width")

    def to_dict(self):
        return {
            "id": self.attachment_id,
            "filename": self.filename,
            "content_type": self.content_type,
            "size": self.size,
            "url": self.url,
            "proxy_url": self.proxy_url,
            "height": self.height,
            "width": self.width,
        }


class Interaction:
    def __init__(self, data: dict):
        self.interaction_id = data.get("id")
        self.interaction_type = data.get("type")
        self.name = data.get("name")
        self.user = User(data.get("user", {}))


class MessageReference:
    def __init__(self, client, ref_data: dict, msg_data: dict):
        self.__client = client
        self.__msg_data = msg_data

        if ref_data:
            self.message_id: Union[str, None] = ref_data.get("message_id")
            self.channel_id: Union[str, None] = ref_data.get("channel_id")
            self.guild_id: Union[str, None] = ref_data.get("guild_id")
            self.fail_if_not_exists: Union[bool, None] = ref_data.get(
                "fail_if_not_exists"
            )
        else:
            self.message_id = (
                self.channel_id
            ) = self.guild_id = self.fail_if_not_exists = None

    @property
    def message(self):
        return Message(self.__client, self.__msg_data or {})

    def to_dict(self):
        return {
            "message_id": self.message_id,
            "channel_id": self.channel_id,
            "guild_id": self.guild_id,
            "fail_if_not_exists": self.fail_if_not_exists,
        }


class Message:
    def __init__(self, client, data: dict):
        self.__data = data
        self.__client = client

        self.message_id: Union[str, None] = data.get("id")
        self.channel_id: Union[str, None] = data.get("channel_id")
        self.guild_id: Union[str, None] = data.get("guild_id")
        self.author: User = User(data.get("author", {}))

        self.member: Member = Member(data.get("member", {}))
        self.content: Union[str, None] = data.get("content")
        self.timestamp: Union[str, None] = data.get("timestamp")
        self.edited_timestamp: Union[str, None] = data.get("edited_timestamp")
        self.tts: Union[bool, None] = data.get("tts")

        self.mention_everyone: Union[bool, None] = data.get("mention_everyone")
        self.mentions: List[User] = [
            User(mention) for mention in data.get("mentions", [])
        ]
        self.mention_roles: List[Role] = [
            Role(role) for role in data.get("mention_roles", [])
        ]
        self.mention_channels: List[ChannelMention] = [
            ChannelMention(client, cm) for cm in data.get("mention_channels", [])
        ]

        self.attachments: List[Attachment] = data.get("attachments") or []
        self.embeds: List[embed.Embed] = data.get("embeds") or []
        self.reactions: List[Reaction] = data.get("reactions") or []
        self.nonce: Union[str, int, None] = data.get("nonce")
        self.pinned: Union[bool, None] = data.get("pinned")
        self.webhook_id: Union[str, None] = data.get("webhook_id")
        self.message_type: Union[int, None] = data.get("type")
        self.activity: MessageActivity = MessageActivity(data.get("activity", {}))

        self.application: Application = Application(data.get("application", {}))
        self.application_id: Union[str, None] = data.get("application_id")
        self.message_reference: MessageReference = MessageReference(self.__client,
            data.get("message_reference", {}), data.get("referenced_message", {})
        )

        self.interaction: Interaction = Interaction(data.get("interaction", {}))
        self.thread = None
        self.components = None
        self.sticker_items: List[StickerItem] = [
            StickerItem(stick) for stick in data.get("sticker_items", [])
        ]

    @classmethod
    def create(
        cls, client,
        *,
        content: str = None,
        reply_to: MessageReference = None,
        embeds: List[embed.Embed] = None,
        attachments: List[Attachment] = None,
    ):
        if reply_to:
            reply = reply_to.to_dict()
        else:
            reply = None

        return Message(client,
            {
                "content": content if content else "",
                "message_reference": reply,
                "embeds": embeds,
                "attachments": attachments,
            }
        )

    @classmethod
    async def get_message(cls, client, channel_id, message_id):
        async with aiohttp.ClientSession() as session:
            url = f"https://discord.com/api/v9/channels/{channel_id}/messages/{message_id}"

            resp = await session.get(
                url,
                headers={"Authorization": f"Bot {client.ws.bot_token}"}
            )
            rs = await resp.json()

            if resp.ok:
                return Message(client, rs)

    async def send(
        self, channel_id: str = None, *, allowed_mentions: dict = None
    ):
        """Send a message to a channel ID, returns a Message object"""
        async with aiohttp.ClientSession() as session:
            url = f"https://discord.com/api/v9/channels/{self.channel_id or channel_id}/messages"

            resp = await session.post(
                url,
                headers={"Authorization": f"Bot {self.__client.ws.bot_token}"},
                json={
                    "content": self.content,
                    "tts": self.tts,
                    "allowed_mentions": allowed_mentions,
                    "embeds": [emb.to_dict() for emb in (self.embeds or [])],
                },
            )
            rs = await resp.json()
            print(rs)
            if resp.ok:
                return Message(self.__client, rs)
            

    async def delete(self):
        async with aiohttp.ClientSession() as session:
            url = f"https://discord.com/api/v9/channels/{self.channel_id}/messages/{self.message_id}"
            await session.delete(
                url, headers={"Authorization": f"Bot {self.__client.ws.bot_token}"}
            )

    # TODO: Implement attachments and files
    async def edit(self, *, content: str = None, 
        embeds: typing.List[models.Embed] = None,
        allowed_mentions: dict = None
    ):
        async with aiohttp.ClientSession() as session:
            url = f"https://discord.com/api/v9/channels/{self.channel_id}/messages/{self.message_id}"
            await session.patch(
                url, headers={"Authorization": f"Bot {self.__client.ws.bot_token}"},
                json={
                    "content": content if content else "",
                    "embeds": [embed.to_dict() for embed in (embeds or [])],
                    "allowed_mentions": allowed_mentions
                }
            )            

    async def pin(self):
        async with aiohttp.ClientSession() as session:
            url = f"https://discord.com/api/v9/channels/{self.channel_id}/pins/{self.message_id}"

            await session.put(
                url, headers={"Authorization": f"Bot {self.__client.ws.bot_token}"}
            )
    
    async def unpin(self):
        async with aiohttp.ClientSession() as session:
            url = f"https://discord.com/api/v9/channels/{self.channel_id}/pins/{self.message_id}"

            await session.delete(
                url, headers={"Authorization": f"Bot {self.__client.ws.bot_token}"}
            )        



import enum
from typing import Dict, List, Union, Any

import aiohttp

import pycordia
from pycordia import models


class Channel:
    class ChannelType(enum.Enum):
        text = 0
        dm = 1
        voice = 2
        group_dm = 3
        category = 4
        news = 5
        store = 6
        news_thread = 10
        public_thread = 11
        private_thread = 12
        stage_voice = 13

    def __init__(self, client, data: dict):
        self.__client = client

        self.name: str = data.get("name")
        self.topic: str = data.get("topic")
        self.id: str = data.get("id")
        self.type: Channel.ChannelType = Channel.ChannelType(int(data.get("type", 0)))

        self.nsfw: bool = data.get("nsfw")
        self.guild_id: Union[str, None] = data.get("guild_id")
        self.position: int = data.get("position")

        # TODO: Define overwrite model
        self.permission_overwrites: List[Dict] = data.get("permission_overwrites")

        self.permissions: str = data.get("permissions")

        # Voice channel attributes
        self.bitrate: Union[int, None] = data.get("bitrate")
        self.user_limit: Union[int, None] = data.get("user_limit")
        self.rtc_region: Union[str, None] = data.get("rtc_region")

        self.slowmode_count: Union[int, None] = data.get("rate_limit_per_user")
        self.recipients: List = data.get("recipients")
        self.icon: Union[str, None] = data.get("icon")

        self.owner_id: Union[str, None] = data.get("owner_id")
        self.application_id: Union[str, None] = data.get("application_id")
        self.parent_id: Union[str, None] = data.get("parent_id")

        self.last_pinned_at: str = data.get("last_pin_timestamp")

        self.message_count: int = data.get("message_count", -1)  # -1 if the count was not provided
        self.member_count: int = data.get("member_count", -1)

        # Thread specific stuff
        # TODO: Define models for these types
        self.thread_metadata: Dict[str, Any] = data.get("thread_metadata", {})
        self.thread_member: Dict[str, Any] = data.get("member", {})
        self.thread_default_auto_archive_duration: int = data.get("default_auto_archive_duration", -1)

    def __repr__(self):
        return f"#{self.name}"

    @property
    def mention(self):
        return f"<#{self.id}>"

    @classmethod
    async def from_id(cls, client, channel_id: str):
        """Get a channel object given the id"""

        async with aiohttp.ClientSession() as session:
            async with session.get(
                f"{pycordia.api_url}/channels/{channel_id}",
                headers={
                    "Authorization": f"Bot {client._Client__bot_token}"
                }
            ) as resp:

                if not resp.ok:
                    return

                return Channel(client, await resp.json())

    # TODO: Make this method support cache
    async def query_message(self, message_id):
        async with aiohttp.ClientSession() as session:
            async with session.get(
                f"{pycordia.api_url}/channels/{self.id}/messages/{message_id}",
                headers={
                    "Authorization": f"Bot {self.__client._Client__bot_token}"
                }
            ) as resp:
                if not resp.ok:
                    return

                return models.Message(self.__client, await resp.json())

class ChannelMention:
    def __init__(self, client, data: dict):
        self.__client = client

        self.channel_id: Union[str, None] = data.get("id")
        self.guild_id: Union[str, None] = data.get("guild_id")
        self.channel_type: Union[str, None] = data.get("type")
        self.channel_name: Union[str, None] = data.get("name")

    def __repr__(self):
        return f"<pycordia.models.ChannelMention - id={self.channel_id} name={self.channel_name}>"

    async def get_channel(self):
        return await Channel.from_id(self.__client, self.channel_id)



# Pycordia v0.1.2 - The Discord bot framework
# Developed by Angel Carias and it's contributors. 2021.

# client.py
#   Handles bot creation


import aiohttp
import asyncio
import enum

import typing

from . import events, models, websocket
import pycordia


class Intents(enum.Enum):
    guilds = 1 << 0
    guild_members = 1 << 1
    guild_bans = 1 << 2
    guild_emojis_and_stickers = 1 << 3
    guild_integrations = 1 << 4
    guild_webhooks = 1 << 5
    guild_invites = 1 << 6
    guild_voice_states = 1 << 7
    guild_presences = 1 << 8
    guild_messages = 1 << 9
    guild_message_reactions = 1 << 10
    guild_message_typing = 1 << 11
    direct_messages = 1 << 12
    direct_message_reactions = 1 << 13
    direct_message_typing = 1 << 14

    @classmethod
    def all(cls):
        """Enables all registered intents, including privileged ones"""
        return cls.merge_intents(cls)
    
    @classmethod
    def merge_intents(cls, intent_list):
        """Convert a list of intents into a number

        ---        
        Parameters:
            intent_list: A list of intents
        """
        result = 0
        for value in intent_list:
            result |= value.value
        return result


class Client:
    """A WebSockets client for the Discord Gateway API"""

    def event(self, fun):
        self.events[fun.__name__] = fun

        def wrapper():
            fun()

        return wrapper

    def __init__(self, intents, cache_size: int = 1000):
        self.events = {}
        self.__bot_token: str = ""
        self.intents = intents

        self.cache_size = int(cache_size)
        self.user_cache: typing.Dict[str, models.User] = {}
        self.message_cache: typing.Dict[str, models.Message] = {}

    async def __create_ws(self, bot_token):        
        self.ws = websocket.DiscordWebSocket(self, bot_token, self.intents)
        await self.ws.listen()

    async def call_event_handler(self, event_name: str, event_data):
        func_name = f"on_{event_name.lower()}"
        
        print(event_name)
        # --- Cached methods ---
        if event_name.lower() == "message_create":
            message = models.Message(self, event_data)

            if len(self.message_cache.keys()) >= self.cache_size:
                first_message = list(self.message_cache.keys())[0]
                del self.message_cache[first_message]

            self.message_cache[message.message_id] = message

            if func_name in self.events:
                asyncio.gather(self.events[func_name](message))                
                return
        elif event_name.lower() == "message_update":
            after = models.Message(self, event_data)
            print(self.message_cache, after.message_id)
            before = self.message_cache.get(after.message_id, None)

            # Update the message cache
            if len(self.message_cache.keys()) >= self.cache_size:
                first_message = list(self.message_cache.keys())[0]
                del self.message_cache[first_message]
            self.message_cache[after.message_id] = after

            if func_name in self.events:
                asyncio.gather(self.events[func_name](before, after))                
                return

        # --- Uncached methods ---
        if func_name in self.events:
            func = self.events[func_name]

            if event_name.lower() == "ready":
                asyncio.gather(func(events.ReadyEvent(event_data)))

            # ---- User Related Events ----
            elif event_name.lower() == "typing_start":
                asyncio.gather(func(events.TypingStartEvent(event_data)))

            # ---- Message Related Events ----
            elif event_name.lower() == "message_delete":
                asyncio.gather(func(
                    events.MessageDeleteEvent(event_data, False, [self.message_cache[event_data["id"]]])
                ))
            elif event_name.lower() == "message_delete_bulk":
                ids = event_data.get("ids", [])
                asyncio.gather(func(
                    events.MessageDeleteEvent(event_data, True, 
                        list(filter(lambda key: key in ids, self.message_cache))
                )))
            
            # ---- Channel Related Events ----
            elif event_name.lower() in ("channel_create", "channel_update", "channel_delete"):
                asyncio.gather(func(models.Channel(self, event_data)))

            # ---- Unimplemented ----
            else:
                asyncio.gather(func(event_data))

    def run(self, bot_token):
        self.__bot_token = bot_token

        loop = asyncio.get_event_loop()
        task = loop.create_task(self.__create_ws(bot_token))

        try:
            loop.run_until_complete(task)
        except KeyboardInterrupt:
            task.cancel()

            if self.ws.sock:
                loop.run_until_complete(self.ws.sock.close())

    @property
    async def guilds(self) -> typing.List[models.PartialGuild]:
        # Else, fetch it from the discord api
        async with aiohttp.ClientSession() as session:
            async with session.get(
                    f"{pycordia.api_url}/users/@me/guilds",
                    headers={
                        "Authorization": f"Bot {self.__bot_token}"
                    }
            ) as resp:

                if not resp.ok:
                    content = await resp.text()
                    raise Exception(content)

                guilds = await resp.json()
                guilds = list(map(models.PartialGuild, guilds))

                return guilds

    @property
    async def user(self) -> models.User:
        "Get user info for the bot"
        async with aiohttp.ClientSession() as session:
            async with session.get(
                    f"{pycordia.api_url}/users/@me",
                    headers={
                        "Authorization": f"Bot {self.__bot_token}"
                    }
            ) as resp:

                if not resp.ok:
                    content = await resp.text()
                    raise Exception(content)

                return models.User(await resp.json())

    # TODO: Modify current user
    # TODO: Create DM
    # TODO: Leave Guild
    # TODO: Get User Connections



"""
### PyCordia - The Discord bot framework

Developed by Angel Carias and it's contributors. 2021.

A wrapper around the Discord HTTP API and WebSockets
"""

from .client import Intents, Client
from . import websocket, errors, events, models

__version__ = "0.1.2"


cdn_url = "https://cdn.discordapp.com"
ws_url = "wss://gateway.discord.gg"
api_url = "https://discord.com/api"



import aiohttp
import pycordia


class User:
    """Model used for users"""
    def __init__(self, json):
        self.user_id: str = json.get("id")
        self.username: str = json.get("username")
        self.discriminator: str = json.get("discriminator")
        self.avatar_hash: str = json.get("avatar")
        self.bot: bool = json.get("bot")
        self.system: bool = json.get("system")
        self.mfa_enabled: bool = json.get("mfa_enabled")
        self.banner: str = json.get("banner")
        self.accent_color: int = json.get("accent_color")
        self.locale: str = json.get("locale")
        self.verified: bool = json.get("verified")
        self.email: str = json.get("email")
        self.flags: int = json.get("flags")
        self.premium_type: int = json.get("premium_type")
        self.public_flags: int = json.get("public_flags")

    def __repr__(self):
        return f"{self.username}#{self.discriminator}"

    @property
    def mention(self):
        return f"<@!{self.user_id}>"

    @property
    def avatar_url(self):
        return f"{pycordia.cdn_url}/avatars/{self.user_id}/{self.avatar_hash}.png"

    @classmethod
    async def user_from_id(cls, client, user_id: str, use_cache: bool = True):
        """Get user info, given the user id"""

        # Check if its present in the cache first -
        user = client.user_cache.get(user_id, None)
        if user and use_cache:
            return user

        # Else, fetch it from the discord api
        async with aiohttp.ClientSession() as session:
            async with session.get(
                f"{pycordia.api_url}/users/{user_id}",
                headers={
                    "Authorization": f"Bot {client._Client__bot_token}"
                }
            ) as resp:

                if not resp.ok:
                    return

                json = await resp.json()
                user = User(json)

                if len(client.user_cache.keys()) >= client.cache_size:
                    first_user = list(client.user_cache.keys())[0]
                    del client.user_cache[first_user]

                client.user_cache[user.user_id] = user

                return user



import typing
import aiohttp
from typing import List, Union, Optional
from .user import User


class Member:
    def __init__(self, data: dict):
        self.user: User = User(data.get("user", {}))
        self.nick: Union[str, None] = data.get("nick")
        self.roles: list = data.get("roles", [])
        self.joined_at: Union[str, None] = data.get("joined_at")
        self.premium_since: Union[str, None] = data.get("premium_since")
        self.deaf: Union[bool, None] = data.get("deaf")
        self.mute: Union[bool, None] = data.get("mute")
        self.pending: Union[bool, None] = data.get("pending")
        self.permissions: Union[str, None] = data.get("permissions")

    def __repr__(self):
        return f"<GuildMember id={self.user.user_id} username='{self.user.username}'>"


class PartialGuild:
    def __init__(self, data: dict):
        self.id: str = data.get("id")
        self.name: str = data.get("name")
        self.icon: str = data.get("icon")
        self.owner: bool = bool(data.get("owner"))
        self.permissions_integer = data.get("permissions")
        self.features: typing.List[str] = data.get("features", [])

    def __repr__(self):
        return f"<PartialGuild - id={self.id} name={self.name}>"


class Role:
    def __init__(self, data: dict):
        self.role_id: Union[str, None] = data.get("id")
        self.role_name: Union[str, None] = data.get("name")

        self.color: Union[int, None] = data.get("color")
        self.colour = self.color

        self.hoist: Union[bool, None] = data.get("hoist")
        self.position: Union[int, None] = data.get("position")
        self.permissions: Union[str, None] = data.get("permissions")
        self.managed: Union[bool, None] = data.get("managed")
        self.mentionable: Union[bool, None] = data.get("mentionable")

        self.tags = data.get("tags")



# Pycordia v0.1.2 - The Discord bot framework
# Developed by Angel Carias and it's contributors. 2021.

# websocket.py
#   Handles communication with the Discord Gateway API


import asyncio
import json
import platform
import zlib
from pycordia.errors import GatewayError

import aiohttp
from aiohttp.client_ws import ClientWebSocketResponse
from aiohttp.http_websocket import WSMsgType

import pycordia

ZLIB_SUFFIX = b"\x00\x00\xff\xff"

class DiscordWebSocket:
    """A WebSockets client for the Discord Gateway API"""

    opcodes = {
        "DISPATCH": 0,
        "HEARTBEAT": 1,
        "IDENTIFY": 2,
        "PRESENCE_UPDATE": 3,
        "VOICE_STATE_UPDATE": 4,
        "RESUME": 6,
        "RECONNECT": 7,
        "REQUEST_GUILD_MEMBERS": 8,
        "INVALID_SESION": 9,
        "HELLO": 10,
        "HEARTBEAT_ACK": 11,
    }

    def __init__(self, client, bot_token, intents):
        self.bot_token = bot_token
        """A valid Discord bot token"""

        self.sock = None

        self.heartbeat_interval = None
        """The interval to wait before calling the gateway again, in milliseconds"""
        self.sequence = None
        self.gateway_url = f"{pycordia.ws_url}/?v=9&encoding=json&compress=zlib-stream"
        self.client = client

        self.session_id = None
        self.intents = intents
        """The provided intents as an integer"""


    def get_identify(self):
        """Returns an Opcode 2 (Identify) response"""
        return {
            "op": self.opcodes["IDENTIFY"],
            "d": {
                "token": self.bot_token,
                "properties": {
                    "$os": platform.system(),
                    "$browser": "pycordia",
                    "$device": "pycordia",
                },
                "compress": True,
                "large_threshold": 250,
                "intents": self.intents,
            },
        }

    async def __keep_alive(self, sock: ClientWebSocketResponse):
        """Keeps the connection alive by sending periodic heartbeats to the Discord gateway"""
        while True:
            if self.heartbeat_interval:
                # Sleep for an interval in seconds
                await asyncio.sleep(self.heartbeat_interval / 1000)

                # Send heartbeat
                await sock.send_str(
                    json.dumps({ "op": self.opcodes["HEARTBEAT"], "d": self.sequence })
                )

    async def __listen_socket(self, sock: ClientWebSocketResponse):
        """Start listening for data from the gateway"""
        self.sock = sock
        inflator = zlib.decompressobj()

        while True:
            data = await sock.receive()
            payload = data.data

            # --- Handle close --- #
            if data.type == WSMsgType.CLOSE:
                code, msg = data.data, data.extra
                await sock.close()
                #  A non-100(0|1) code indicates an error
                if code not in (1000, 1001):
                    raise GatewayError(code, msg)
            # --- Handle binary response --- #
            elif data.type == WSMsgType.BINARY:
                # Decompress the binary into readable JSON data
                buffer = bytearray()
                buffer.extend(data.data)
                # We have to account for both payload and transport
                # compression
                if buffer[-4:] == ZLIB_SUFFIX:
                    payload = inflator.decompress(buffer)
                else:
                    payload = zlib.decompress(buffer)

            if data.type in (WSMsgType.BINARY, WSMsgType.TEXT):
                js = json.loads(payload)
                event_data = js["d"]

                self.sequence = js.get("s")

                # Send identify
                if js["op"] == self.opcodes["HELLO"]:
                    self.heartbeat_interval = event_data["heartbeat_interval"]
                    ident = self.get_identify()
                    await sock.send_json(ident)
                # Call event handlers
                elif js["op"] == self.opcodes["DISPATCH"]:
                    if js["t"] == "READY":
                        self.session_id = event_data["session_id"]

                    await self.client.call_event_handler(js["t"], event_data)

    async def listen(self):
        """Manage a separate connection for both heartbeats and websocket events"""
        async with aiohttp.ClientSession() as session:
            sock = await session.ws_connect(self.gateway_url)

            await asyncio.gather(
                asyncio.create_task(self.__listen_socket(sock)),
                asyncio.create_task(self.__keep_alive(sock)),
            )


import typing
import inspect

def obj_from_dict(data: dict, obj: typing.Callable, alias: dict = None):
    """Convert a dictionary `data` into an `obj`.
    
    Parameters:
        data (dict): The data to provide.
        obj (typing.Callable): The object created from `data`
        alias (dict): A dictionary containing aliases for any parts of the dictionary. \
            Follows a (param-alias) pair
    """
    sig = inspect.signature(obj)
    new_inf = {}
    for param in sig.parameters:
        if alias and param in alias:
            new_inf[param] = data[alias[param]]
        elif param in data:    
            new_inf[param] = data[param]

    return obj(**new_inf)

def obj_to_dict(obj: typing.Any, alias: dict = None):
    """Convert an `obj` into a dictionary.
    
    Parameters:
        obj (typing.Any): The object to convert
        alias (dict): A dictionary containing aliases for any parts of the object. \
            Follows a (attribute-alias) pair
    """
    new_inf = {}
    if obj.__dict__:
        for key, value in obj.__dict__.items():
            if not key.startswith("_"):
                if alias and key in alias:
                    new_inf[alias[key]] = value
                else:
                    new_inf[key] = value

    return new_inf        



class QueryError(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message
        super().__init__(f"{message} (error code {code})")

class GatewayError(QueryError):    
    pass

class ComponentError(Exception):
    pass


from setuptools import setup
import re

requirements = [
    "aiohttp"
]

with open('pycordia/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE)
    if version:
        version = version.group(1)

with open('README.md', "r", encoding="utf-8-sig") as f:
    readme = f.read()

setup(name='pycordia',
      author='Angel Carias',
      url='https://github.com/angelCarias/pycordia',
      version=version,
      packages=["pycordia", "pycordia/models"],
      license='MIT',
      description='A wrapper around the Discord HTTP API and WebSockets',
      long_description=readme,
      long_description_content_type="text/markdown",
      install_requires=requirements,
      python_requires='>=3.7.0',
      classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9"
      ]
)
